Unit FWMath3;
{----------------------------------------------}
{-                            -}
{-                 * FWMATH3 *                -}
{-    Copyright   .., 1992-2001     -}
{----------------------------------------------}
interface
type
  TfmMath3 = class(TObject)
  private
    { Private declarations }
    Procedure ChangeClNmb;
    Function MakeDistMatr: boolean;
    Function Normir: boolean;
    Procedure InitArrays;
    Procedure DelArrays;
    Procedure SaveData;
    Procedure RestoreData;
    Procedure PrintClust(l,dd,aii,ajj: integer);
    Procedure MakeCoeff(var al,be,ga,de: double;
                        aii,ajj,l: integer;
                        SuppleCf: double);
    Function Hierarh(var StepWellClust: integer;
                      SuppleCf: double): boolean;
    Procedure LastStep(var mm: integer;
                       var dd: integer);
    Procedure MakeClustArray(ajj,aii,l: integer;
                           d_min1: double);
    Procedure StatClust(Clb: array of integer);
    Procedure ObjectOut;
    Procedure AddCluster;
  public
    { Public declarations }
    Function Cluster:boolean;
end;

var fmMath3: TfmMath3;
{--------------------------------------------------------}
implementation
uses FWData, FWStat0, FWArray, FWMath1, FWTitle,
     Windows, Dialogs, SysUtils, System.UITypes;

{----------   -----------}
Procedure TfmMath3.InitArrays;
var
   j: integer;
begin
    SetLength(d10,mSize,mSize);
    SetLength(Kn,(2*mSize));
    SetLength(Cln,mSize);
    SetLength(Clb,mSize);
    SetLength(Cl,mSize); // Cl   ,  .
                         //    .
    for j:=0 to m1 do
      begin
        Cln[j]:=j+1;      {     }
        Kn[j]:=1;         {     - 1,   }
        Kn[j+m1]:=0;      {    - 0 }
      end;
end;

Procedure TfmMath3.DelArrays;
begin
   d10:=nil;
   Kn:=nil;
   Cln:=nil;
   Clb:=nil;
   Cl:=nil;
end;

{----------    ---------}
Procedure TfmMath3.LastStep(var mm: integer;
                            var dd: integer);
begin
   if m1<=10
          then
              begin
                 mm:=m1-1;
                 dd:=1;
              end
          else
              begin
                 mm:=10;  {   <=  10}
                 dd:=m1-10;
              end;
end;

{---------- (    MakeDistMatr)------}
Function TfmMath3.Normir;
//     
var
   mean,sigma,sum1,sum2,disp: double;
   i,j: integer;
begin
     for i:=0 to n1 do
       begin
          sum1:=0.0;
          for j:=0 to m1 do
            sum1:=sum1+a[i,j];
          mean:=sum1/(m1+1);

          sum2:=0.0;
          for j:=0 to m1 do
            sum2:=sum2+sqr(a[i,j]-mean);
          disp:=sum2/m1;
          sigma:=sqrt(disp);

          if sigma=0 then
                 begin
                    MessageDlg(' : '+Trim(Hd[i])+'  @=0',mtError,[mbOk],0);
                    Result:=false;
                    Exit;
                 end;

          for j:=0 to m1 do
            a[i,j]:=(a[i,j]-mean)/sigma;
     end;
     Result:=true;
end;

{---------    -------}
{         }
Procedure TfmMath3.SaveData;
var
   i,j: integer;
   fg: text;
begin
   assign(fg,'data.tmp');
   rewrite(fg);
   for j:=0 to m1 do    {    }
     begin
        for i:=0 to n1 do
          write(fg,a[i,j]:9:2,' ');
        writeln(fg);
     end;
   close(fg);
end;

Procedure TfmMath3.RestoreData;
var
   i,j: integer;
   fg: text;
begin
   assign(fg,'data.tmp');
   reset(fg);
   for j:=0 to m1 do    {     }
     begin
        for i:=0 to n1 do
          read(fg,a[i,j]);
        readln(fg);
     end;
   close(fg);
   Erase(fg);
end;

{---------    ----------}
Function TfmMath3.MakeDistMatr;
var
   k: byte;
   i,j,jk: integer;
   sum: double;
   amin,amax: double;
   rs: array of double;
   a1: array of double;

begin
   fmMain.pnlInform.Caption:='  ...';
   fmMain.pnlInform.Refresh;
   SetLength(rs,nSize);
   SetLength(a1,mSize);

   case ObjDistance of
      0: begin     {   }
            SaveData;
            if not Normir
               then
                   begin
                      Result:=false;
                      rs:=nil;
                      a1:=nil;
                      RestoreData;
                      Exit;
                   end;
         end;
      1: begin     {   }
           for k:=0 to n1 do
             begin
                for jk:=0 to m1 do
                  a1[jk]:=a[k,jk];
                fmMath1.MinMax(a1,m1,amin,amax);
                rs[k]:=amax-amin;
                if rs[k]=0 then
                   begin
                      MessageDlg(Trim(Hd[k])+':   (max-min=0)',mtError,[mbOk],0);
                      Result:=false;
                      rs:=nil;
                      a1:=nil;
                      Exit;
                   end;
             end;
         end;
   end;

   for i:=0 to m1 do
     begin
         Inc(nm);
         fmMain.ProgressBar1.Position:=nm;
         for j:=i+1 to m1 do
           begin
              sum:=0.0;       { }
              case ObjDistance of
                0: begin      { }
                      for k:=0 to n1 do
                        sum:=sum+sqr(a[k,i]-a[k,j]); // , 
                      d10[i,j]:=sqrt(sum);       // d10[i,j]  
                   end;
                1: begin      { }
                      for k:=0 to n1 do
                        sum:=sum+abs(a[k,j]-a[k,i])/rs[k];
                      d10[i,j]:=sum;
                   end;
                2: begin      { }
                      for k:=0 to n1 do
                        if a[k,j]<>a[k,i]
                           then d10[i,j]:=d10[i,j]+1;
                   end;
              end;
              d10[j,i]:=d10[i,j]; //..    
           end;                   //    
     end;

   if IsDistance then
     begin
         writeln(g);
         writeln(g,'   ');
         writeln(g,'__________________________________');
         write(g,'      ');
         for i:=1 to m1 do
           write(g,i:5,' ');
         writeln(g);

         for i:=1 to m1 do
           begin
              write(g,(i+1):5,' ');
              for j:=0 to m1 do
                if j<i then
                           write(g,d10[i,j]:5:2,' ')
                       else
                           write(g,'      ');
              writeln(g);
           end;
     end;

   fmMain.pnlInform.Caption:='';
   fmMain.pnlInform.Refresh;
   rs:=nil;
   a1:=nil;
   if ObjDistance=0
      then RestoreData;
   Result:=true;
end;

{-----     1 ----}
Procedure TfmMath3.ChangeClNmb;
var
  j,k,l,z: integer;
begin
   for l:=1 to ClustNumber do
     begin
        k:=0;
        z:=-99;
        repeat
           z:=Clb[k];
           Inc(k);
        until (z>0) or (k>m1);

        for j:=0 to m1 do
          if Clb[j]=z then clb[j]:=l*(-1); // . 
     end;

   for j:=0 to m1 do
     Clb[j]:=abs(Clb[j]);
end;

{-------    -------}
Procedure TfmMath3.MakeClustArray(ajj,aii,l: integer;
                                  d_min1: double);
var
   c1,c2: double;
begin
   with Cl[l] do
      begin
         clevel:=d_min1;
         if (aii>m1) and (ajj>m1)
               then
                  begin
                     c1:=Cl[aii-m1].clevel;  {     }
                     c2:=Cl[ajj-m1].clevel;  {      }
                     if c1>=c2
                        then
                           begin
                              iclson:= ajj;
                              icrson:= aii;
                           end
                        else
                           begin
                              iclson:= aii;
                              icrson:= ajj;
                           end;
                  end
              else
                  if ajj<aii  {   ajj  aii >m }
                     then
                        begin
                            iclson:= aii;
                            icrson:= ajj;
                        end
                     else
                        begin
                            iclson:= ajj;
                            icrson:= aii;
                        end;
         newclust:=m1+l+2;
     end;
end;

{--------    ------}
Procedure TfmMath3.PrintClust(l,dd,aii,ajj: integer);
var
   k,j,ncl: integer;
   dcl: extended;
begin
   with Cl[l] do
     begin
        if l=0 then
              begin
                 writeln(g);
                 writeln(g,'   : ',curr_obj_dist);
                 writeln(g,'  : ',curr_clust_dist);
              end;
                            {    }
        if (l>dd) and (l<m1) then
           begin
               writeln(g);
               writeln(g,'_____________________________');
               writeln(g,' ',l);
               writeln(g,' : ',aii,'  ',ajj);
               writeln(g,'  : ',m1-l);
               writeln(g,' : ',clevel:-6:2);
               dcl:=Cl[l].clevel-Cl[l-1].clevel;
               writeln(g,'    : ',dcl:-6:2);
               writeln(g);

               if l<m1-1 then
                 begin
                    writeln(g,'','     ');
                    for j:=1 to (m1+l+2) do   {  }
                      begin
                         ncl:=1;
                         for k:=0 to m1 do
                           begin
                              if Cln[k]=j then
                                      begin
                                         if ncl=1 then
                                               write(g,'N=',j:3,':      ');
                                         write(g,(k+1),' ');
                                         if frac(ncl/20)=0.0 then
                                              begin
                                                 writeln(g);
                                                 write(g,'            ');
                                              end;
                                         Inc(ncl);
                                      end;
                           end;
                         if ncl>1 then writeln(g)
                      end;
                  end;
           end;
     end; {with}
end;

{-     -}
Procedure TfmMath3.MakeCoeff(var al,be,ga,de: double;
                             aii,ajj,l: integer;
                             SuppleCf: double);
begin
  with Cl[l] do
   begin
     case ClusterDistance of
       0: begin              { " " }
             al:=0.5;
             be:=0.5;
             ga:=0.0;
             de:=-0.5;
          end;
       1: begin              { " " }
             al:=0.5;
             be:=0.5;
             ga:=0.0;
             de:=0.5;
          end;
       2: begin              { " " }
             al:=Kn[ajj-1]/Kn[newclust-1];
             be:=Kn[aii-1]/Kn[newclust-1];
             ga:=0.0;
             de:=0.0;
          end;
       3: begin              { " " }
             al:=Kn[ajj-1]/Kn[newclust-1];
             be:=Kn[aii-1]/Kn[newclust-1];
             ga:=(Kn[ajj-1]*Kn[aii-1])/sqr(Kn[newclust-1]);
             de:=0.0;
          end;
       4:;                   {   - .    }
       5: begin              { " "}
             ga:=SuppleCf;
             al:=(1-ga)/2;
             be:=al;
             de:=0.0;
          end;
     end;
   end;
end;

{-------   -------}
Function TfmMath3.Hierarh(var StepWellClust: integer;
                           SuppleCf: double): boolean;
var
   dd,x,i,j,k,l,mm,ih,ai,aj,aii,ajj: integer;
   c1,c2,c3,d_min,dl,denomin,amin,amax,th: double;
   al,be,ga,de: double;
   a1: array of double;
   a2: array of integer;
begin
 try
   m1:=mx[1]-1;
   nm:=0;               //  
   fmMain.ProgressBar1.Min:=0;
   fmMain.ProgressBar1.Max:=m1*2-1;
   SetLength(a1,mSize); // 
   SetLength(a2,mSize); // 1    (0=>1, 1=1)
   InitArrays;
   LastStep(mm,dd);
   if not MakeDistMatr
       then
           begin
              Result:=false;
              a1:=nil;
              a2:=nil;
              DelArrays;
              Exit;
           end;

   fmMain.pnlInform.Caption:='  ...';
   fmMain.pnlInform.Refresh;
   //  ml,   
   for l:=0 to (m1-1) do
     begin
       Inc(nm);
       fmMain.ProgressBar1.Position:=nm;
       ai:=0;
       aj:=1;
       d_min:=d10[ai,aj];
       for i:=0 to m1 do            {   }
         for j:=(i+1) to m1 do
           begin
            if d10[i,j]<d_min then
              begin
                d_min:=d10[i,j];    { min  }
                ai:=i;
                aj:=j;
              end;
           end;
       aii:=Cln[ai];            {  1-   }
       ajj:=Cln[aj];            {  2-   }
       MakeClustArray(ajj,aii,l,d_min);

       with Cl[l] do
         begin
           for i:=0 to m1 do         {   }
             if (Cln[i]=ajj) or (Cln[i]=aii) {     }
                then
                   begin
                      Cln[i]:=newclust;
                      Kn[newclust-1]:=Kn[newclust-1]+1;
                   end;

           if ((Kn[ajj-1]=1) or (Kn[aii-1]=1)) and (l>1)
             then a2[l-1]:=1;      {    n=1}

           if l=(m1-ClustNumber) //    
             then                //   ,
              begin              //      
                 for j:=0 to m1 do
                   Clb[j]:=Cln[j];
                 ChangeClNmb;
              end;

           PrintClust(l,dd,aii,ajj);
           MakeCoeff(al,be,ga,de,aii,ajj,l,SuppleCf);

           dl:=d10[ai,aj];
           for i:=0 to m1 do        {    }
             begin                  { Lance G.N.,Williams W.T.,1967}
                for j:=(i+1) to m1 do
                  begin
                    if ClusterDistance=4 then  {  }
                      begin
                         if (j=ai) or (i=ai)
                           then
                              begin
                                 if j=ai then x:=Cln[i];
                                 if i=ai then x:=Cln[j];
                                 denomin:=Kn[x-1]+Kn[aii-1]+Kn[ajj-1];
                                 al:=(Kn[x-1]+Kn[aii-1])/denomin;
                                 be:=(Kn[x-1]+Kn[ajj-1])/denomin;
                                 ga:=-Kn[x-1]/denomin;
                                 de:=0.0;
                               end;
                      end;

                    if j=ai then               { }
                      d10[i,j]:=al*d10[i,ai]+be*d10[i,aj]+ga*dl+
                            de*abs(d10[i,ai]-d10[i,aj]);
                    if i=ai then               { }
                      d10[i,j]:=al*d10[ai,j]+be*d10[aj,j]+ga*dl+
                            de*abs(d10[ai,j]-d10[aj,j]);
                                    {""  }
                    if (aj=i) or (aj=j) then d10[i,j]:=1e7;
                  end;
                                    {   }
                for j:=0 to i do
                  d10[i,j]:=d10[j,i];
             end;
         end; { with }
     end; { m1 }

   Cl[m1].clevel:=Cl[m1-1].clevel; {=    }
   for i:=0 to (m1-1) do
     a1[i]:=Cl[i+1].clevel-Cl[i].clevel;     { }
   fmMath1.MinMax(a1,(m1-1),amin,amax);
   th:=amax;              {   }

   StepWellClust:=255;
   if (m1-mm)>1
              then
                  ih:=m1-mm
              else
                  ih:=2;

   for i:=ih to (m1-1) do
     begin
        if (a2[i]<>1) and (th=a1[i])
           then      {      +  1  }
                     {      }
               begin
                  IsStepWellClust:=true;
                  StepWellClust:=i;
                  writeln(g,'     ',i);
                  writeln(g,'  ',m1-i,' ');
                  a2[i]:=2;
               end;
     end;

   writeln(g);
   writeln(g);
   writeln(g,'  ');
   writeln(g,'_____________________________');
   writeln(g,'     ');
   for i:=(m1-mm) to (m1-2) do
     begin
        if a2[i+1]=2 then
                       write(g,'! ')
                   else
                       write(g,'  ');

        write(g,(i+1):2,'  ',a1[i]:6:3,' ');

        j:=round(a1[i]*60/th);
        for k:=1 to j do
          write(g,'*');
        writeln(g);
     end;
   Result:=true;
  finally
    d10:=nil;
    Kn:=nil;
    Cln:=nil;
    a1:=nil;
    a2:=nil;
    writeln(g);
    fmMain.pnlInform.Caption:='';
    fmMain.pnlInform.Refresh;
    fmMain.ProgressBar1.Position:=0;
  end;
end;

{----------------------------------------------------------------}
Procedure TfmMath3.StatClust(Clb: array of integer);
var
   bd,i,j,k,l,mn,z: integer;
   sum1,sum2,disp: double;
   mean1,sigma1,mm1: double;
begin
   writeln(g); writeln(g);
   writeln(g,' ');
   writeln(g,'__________ ');
   z:=-1;
   for i:=0 to m1 do
     begin
      if clb[i]>0 then
       begin
        bd:=clb[i];
        for j:=i to m1 do
          if clb[j]=bd             {   }
            then clb[j]:=z;

        writeln(g,' ',bd);
        fmTitle.SubTitle1;
        for l:=0 to n00 do
          begin
             sum1:=0;
             mn:=0;
             for k:=0 to m1 do
               if clb[k]=z then
                 begin
                    sum1:=sum1+a[l,k];
                    Inc(mn);
                 end;
             if mn>1 then
                        begin
                           mean1:=sum1/mn;
                           sum2:=0.0;
                           for k:=0 to m1 do
                             if clb[k]=z then
                                 sum2:=sum2+sqr(a[l,k]-mean1);
                           disp:=sum2/(mn-1);
                           sigma1:=sqrt(disp);
                           mm1:=sigma1/sqrt(mn-1);
                           writeln(g,'   ',Hd[l]:8,'     ',mn,'   ',mean1:8:3,'  ', sigma1:9:3,'  ',mm1:8:3);
                        end
                     else
                        begin
                           writeln(g,'   ',Hd[l]:8,'     ',mn,'   ',sum1:8:3,'      ', '-----','     ','-----');
                        end;
          end;
         writeln(g);
         Dec(z);
       end;
     end;
end;

{-------------------------------------------------------}
Procedure TfmMath3.ObjectOut;
var
   i,j,k,l: integer;
begin
   writeln(g);writeln(g);
   writeln(g,'      ');
   writeln(g,' ___________________________________ ');
     for i:=0 to m1 do
       begin
          l:=0;
          for j:=0 to m1 do
            begin
               if (i+1)=clb[j] then
                 begin
                    Inc(l);
                    writeln(g);
                    if l=1 then
                               begin
                                  writeln(g,' ',(i+1));
                                  write(g,'   N   ');
                                  for k:=0 to n00 do
                                    write(g,Hd[k]:8);
                                  writeln(g);
                               end;
                    write(g,(j+1):4,'  ');

                    for k:=0 to n00 do         {   }
                      write(g,a[k,j]:8:3,' ');
                 end;
            end;
       end;
end;

{-------------------------------------------------------}
Procedure TfmMath3.AddCluster;
var
   amax,amin,j,nSize0: integer;
   amax_double: double;
   ns: string[2];
begin
   Inc(cl_n);
   Inc(n00);
   nSize0:=n00+1;
   Str(cl_n,ns);
   SetLength(Hd,nSize0);
   Hd[n00]:='CL_'+ns;
   SetLength(a,nSize0,mSize);
   for j:=0 to m1 do
     a[n00,j]:=Clb[j];
   SetLength(mx,nSize0);
   mx[n00]:=mSize;
   fmMath1.MinMaxInt(Clb,m1,amin,amax);
   amax_double:=amax;
   SetLength(StepYarray,nSize0);
   case SetYscale of
     0,1: StepYarray[n00]:=fmArray.ScaleSelect(amax_double);     {  }
     2: StepYarray[n00]:=(StepY*amax_double)/StandWinHeght;{     }
   end;
end;

Function TfmMath3.Cluster;
  {-----------------------------------------------------------------}
  {-      : 17.08.95                                           -}
  {-                                                               -}
  {-      :  pp p              -}
  {-                                                               -}
  {-      : g -                             -}
  {-            a - p                  -}
  {-            n -   pp                         -}
  {-           nl -  p pp  pp   -}
  {-           mx -  p ()                       -}
  {-           Hd -   pp                 -}
  {-           Cl -  p  p                 -}
  {-                                                               -}
  {-      : Clb -  p p      -}
  {-            clustnmb -                    -}
  {-            ,n,mx,Hd -    1, ..    -}
  {-                        pp p p             -}
  {-                                                               -}
  {-  :                                                  -}
  {-  1.  .., ..          -}
  {-          .-.:,1983.-   -}
  {-       208 .                                                  -}
  {-  2.  .., ..   -   -}
  {-       .- .:   / .     -}
  {-       .  .: .  .-.:,1986,.269-      -}
  {-       300.                                                    -}
  {-                                                               -}
  {-  :             -}
  {-     IMSL,               -}
  {-  Statgraphics!                                                -}
  {-----------------------------------------------------------------}
begin
   fmTitle.SubTitle13;
   IsStepWellClust:=false;
   if Hierarh(StepWellClust,ClustCf)
       then
           begin
              StatClust(clb);
              if IsClustCont
                then ObjectOut;
              if IsAddClust
                then AddCluster;
              Result:=true;
           end
       else Result:=false;
end;
end.
